//	$Id: dlgdata.cpp,v 1.42 2011/01/07 16:08:38 sugiura Exp $
/*
 *	dlgdata.cpp
 *	_CAONX
 */

#include "session.h"
#include "dlgdata.h"
#include "stack.h"
#include "ddfile.h"
#include "ctrldata.h"
#include "misc.h"
#include "str_tbl.h"

const StringBuffer strRootPageName("root");
const StringBuffer okStr("OK");
const StringBuffer ngStr("NG");

//	VXetHg(DEFAULT_GUI_FONT)̑𓾂
BOOL
GetDefFontProperty(LPFontProperty pfp)
{
	if (pfp == NULL) return FALSE;
	HDC	hDc = ::GetDC(::GetDesktopWindow());
	if (hDc == NULL) return FALSE;
	HGDIOBJ	hOrgFont = ::SelectObject(hDc,::GetStockObject(DEFAULT_GUI_FONT));
	::GetTextFace(hDc,32,pfp->m_fontname.getBufPtr());
	TEXTMETRIC tm;
	::GetTextMetrics(hDc,&tm);
	pfp->m_fontsize = ::MulDiv(tm.tmHeight,72,::GetDeviceCaps(hDc,LOGPIXELSY));
	::SelectObject(hDc,hOrgFont);
	::ReleaseDC(::GetDesktopWindow(),hDc);
	return TRUE;
}

static POINTS
GetDlgPosOrigin(WORD wPosOrigin, HWND hwndFrame)
{
	HWND hwndParent = ::GetParent(hwndFrame);

	RECT crect, drect;
	::GetWindowRect(hwndFrame, &drect);
	::GetClientRect(hwndParent, &crect);
	POINT cs = { 0, 0 };
	::ClientToScreen(hwndParent, &cs);
	crect.left += cs.x;  crect.right  += cs.x;
	crect.top  += cs.y;  crect.bottom += cs.y;

	POINT pos;
	switch (wPosOrigin) {
	case DLGPOS_SCREEN_CENTER:
		/*	Screen Center	*/
		::GetWindowRect(::GetDesktopWindow(), &crect);
		pos.x = ((crect.left + crect.right) >> 1)
				- ((drect.right - drect.left) >> 1);
		pos.y = ((crect.top + crect.bottom) >> 1)
				- ((drect.bottom - drect.top) >> 1);
		break;
	case DLGPOS_CLIENT_COORD:
		/*	Window Coordinate	*/
		pos.x = crect.left;
		pos.y = crect.top;
		break;
	case DLGPOS_SCREEN_COORD:
		pos.x = pos.y = 0;
		break;
	case DLGPOS_CARET_COORD:
		/*  Caret Coordinate  */
		::GetCaretPos(&pos);
		::ClientToScreen(hwndParent, &pos);
		break;
	case DLGPOS_CURSOR_COORD:
		/*  Cursor Coordinate  */
		::GetCursorPos(&pos);
		break;
	case DLGPOS_CLIENT_CENTER:
	default:
		pos.x = ((crect.left + crect.right) >> 1)
				- ((drect.right - drect.left) >> 1);
		pos.y = ((crect.top + crect.bottom) >> 1)
				- ((drect.bottom - drect.top) >> 1);
		break;
	}

	POINTS rpos;
	rpos.x = (WORD)pos.x;
	rpos.y = (WORD)pos.y;

	return rpos;
}

static POINTS
GetDlgPosByPixel(const POINTS& pos, WORD wPosUnit, HWND hwndFrame)
{
	DWORD dlgbaseunit = GetDialogBaseUnits(hwndFrame, "M");
	POINTS ppos;
	if (wPosUnit == DLGPOS_UNIT_FONTSIZE) {
		ppos.x = (pos.x * LOWORD(dlgbaseunit)) >> 2;
		ppos.y = (pos.y * HIWORD(dlgbaseunit)) >> 3;
	} else {
		// sNZP
		ppos = pos;
	}

	return ppos;
}

static POINTS
GetDlgPosByUnit(const POINTS& pos, WORD wPosUnit, HWND hwndFrame)
{
	DWORD dlgbaseunit = GetDialogBaseUnits(hwndFrame, "M");
	POINTS rpos;
	if (wPosUnit == DLGPOS_UNIT_FONTSIZE) {
		rpos.x = pos.x * 4 / LOWORD(dlgbaseunit);
		rpos.y = pos.y * 8 / HIWORD(dlgbaseunit);
	} else {
		// sNZP
		rpos = pos;
	}

	return rpos;
}

//	DLGTEMPLATE ̍\z
WORD*
CreateDialogTemplate(
	LPDLGTEMPLATE pDlgTemplate,
	DWORD style, DWORD exstyle,
	WORD cdit,
	WORD x, WORD y, WORD cx, WORD cy,
	const StringBuffer& title,
	const StringBuffer& fontname, WORD fontsize)
{
	//	DLGTEMPLATE
	pDlgTemplate->style				= style;
	pDlgTemplate->dwExtendedStyle	= exstyle;
	pDlgTemplate->cdit				= cdit;
	pDlgTemplate->x					= x;
	pDlgTemplate->y					= y;
	pDlgTemplate->cx				= cx;
	pDlgTemplate->cy				= cy;

	LPWORD lpwExData = reinterpret_cast<LPWORD>(pDlgTemplate + 1);
	lpwExData += 2;	//	j[ANX͂Ȃ
	lpwExData += title.toUnicode((LPWSTR)lpwExData);

	*lpwExData++ = fontsize;
	lpwExData += fontname.toUnicode((LPWSTR)lpwExData);

	//	DLGITEMTEMPLATE  DWORD En܂Kv!!
//	if ((reinterpret_cast<DWORD>(lpwExData))%sizeof(DWORD)) lpwExData++;
	if (reinterpret_cast<DWORD>(lpwExData) & (sizeof(DWORD) - 1)) lpwExData++;

	return lpwExData;
}


//	q_CAÕR[obN֐
BOOL CALLBACK
DlgPageProc(HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
	DlgPage* pdp;
	HWND hwndFrame;

	if (uMsg == WM_INITDIALOG) {
		//	q_CAȌ
		pdp = reinterpret_cast<DlgPage*>(lParam);
		if (!pdp->initPage(hDlg)) {
			::PostMessage(pdp->getDlgFrame().getUserDlg(),WM_CLOSE,0,0);
		}
		::SetWindowLongPtr(hDlg, DWLP_USER, (LONG_PTR)lParam);
		return FALSE;
	}

	//	q_CAONXւ̃|C^̎擾
	if ((pdp = reinterpret_cast<DlgPage*>(::GetWindowLongPtr(hDlg,DWLP_USER)))
		== NULL) return FALSE;

	switch (uMsg) {
	case WM_USER_MESSAGEBOX:
	case WM_USER_QUESTIONBOX:
		{
			//	bZ[W{bNX̏͐e_CAOɔC
			HWND hwndFrame = pdp->getDlgFrame().getUserDlg();
			if (hwndFrame != NULL) ::PostMessage(hwndFrame,uMsg,wParam,lParam);
		}
		break;

	case WM_COMMAND:
	case WM_NOTIFY:
	case WM_HSCROLL:
		{
//			::OutputDebugString("Enter DlgPage WM_COMMAND(1)\n");
			WORD notify = 0xFFFF;
			switch (uMsg) {
			case WM_COMMAND:
//				::OutputDebugString("Enter DlgPage WM_COMMAND(2)\n");
				notify = pdp->onCommand(wParam,lParam);
				break;
			case WM_NOTIFY:
//				::OutputDebugString("Enter DlgPage WM_COMMAND(3)\n");
				notify = pdp->onNotify(wParam,lParam);
				break;
			case WM_HSCROLL:	//	gbNo[̒ʒmR[h擾
//				::OutputDebugString("Enter DlgPage WM_COMMAND(4)\n");
				notify = pdp->onHScroll(wParam,lParam);
				break;
			}
			if (notify != 0xFFFF) {
				//	ʒmR[h̑M
//				::OutputDebugString("Enter DlgPage WM_COMMAND(5)\n");
				hwndFrame = pdp->getDlgFrame().getUserDlg();
				if (hwndFrame != NULL)
					::PostMessage(hwndFrame,WM_USER_NOTIFY,(WPARAM)notify,0L);
			}
		}
		break;

	case WM_CTLCOLORDLG:
		{
			HBRUSH	hBrush = pdp->getBkBrush();
			const StringBuffer& name = pdp->getName();
			DEBUG_OUTPUT(("WM_CTLCOLORDLG is posted to DlgPage(%s), returns 0x%08x", (LPCSTR)name, (DWORD)hBrush));
			return (BOOL)hBrush;
		}

	case WM_CTLCOLORSTATIC:
	case WM_CTLCOLORBTN:
	case WM_CTLCOLORLISTBOX:
	case WM_CTLCOLOREDIT:
		//	Rg[̃tHg̐FύX
		return reinterpret_cast<BOOL>(pdp->onCtlColor(uMsg, wParam, lParam));
//		return FALSE;

#if 0
	case WM_CTLCOLORSTATIC:
		pdp->onCtlColor(uMsg, wParam, lParam);
		return FALSE;
#endif

	case DM_SETDEFID:
		// prevent the default message handling
		break;

	case WM_CLOSE:
		DEBUG_OUTPUT(("Enter WM_CLOSE in DlgPage"));
		hwndFrame = pdp->getDlgFrame().getUserDlg();
		if (hwndFrame != NULL) ::PostMessage(hwndFrame, WM_CLOSE, 0, 0);
		else ::DestroyWindow(hDlg);
		break;

	case WM_DESTROY:
		// this is invoked in uninitFrame()
		// pdp->uninitPage();
		break;

	default:
		return FALSE;
	}
	return TRUE;
}

DlgPage::DlgPage(DlgFrame* pdf, const StringBuffer& name, WORD width)
	:	m_pDlgFrame(pdf),
		m_strName(name),
		m_hwndPage(NULL),
		m_dwPageSize(0),
		m_pCtrlList(new LinkList<CtrlListItem>),
		m_pHashCtrl(new CtrlHashTable),
		m_pCurCtrlData(NULL),
		m_pShowPageQueue(new LinkList<ShowPageQueueItem>),
		m_bVisible(FALSE),
		m_bEnable(TRUE),
		m_cdit(0),
		m_iniwidth(width),
		m_width(0),
		m_height(0),
		m_crefBkColor(COLOR_3DFACE),
		m_hbkBrush(NULL)
{
	//	nothing to do more.
}

DlgPage::~DlgPage()
{
	if (m_hwndPage != NULL) ::DestroyWindow(m_hwndPage);
	delete m_pCtrlList;
	delete m_pHashCtrl;
	delete m_pShowPageQueue;
}

//	q_CAÕEBhE̐
HWND
DlgPage::createPage(
	HWND hwndCtrl,
	BYTE hbId,
	WORD x, WORD y, WORD cx, WORD cy,
	bool bInTabCtrl,
	FontProperty& fp)
{
	//	hwndCtrl	:	q_CAORg[̃EBhEnh
	//					NULL łȂꍇAx, y ͂̃Rg[̍ォ
	//					ItZbgɂȂ

	if (m_dwPageSize == 0) {
		//	q_CAÕev[g̃TCYƃEBhȆ傫̕]
		if (!this->evalPageSize()) return FALSE;
	}

	//	DLGTEMPLATE ̑傫𑫂
	m_dwPageSize += sizeof(DLGTEMPLATE) + 
//					sizeof(WORD) * (fp.m_fontname.length() + 1) + 16;
					((fp.m_fontname.length() + 1) << 1) + 16;
	//	obt@̊m
	Array<BYTE> pDlgTemplate(m_dwPageSize);
	pDlgTemplate.zero();

	//	DLGTEMPLATE ̍\z
	WORD* lpwExData = CreateDialogTemplate((LPDLGTEMPLATE)(BYTE*)pDlgTemplate,
										WS_CHILD|WS_VISIBLE|
											WS_CLIPCHILDREN|WS_CLIPSIBLINGS|
											DS_CONTROL|DS_SETFONT,
										0L,
										m_cdit,
										x, y, cx, cy,
										nullStr,
										fp.m_fontname,
										(WORD)fp.m_fontsize);

	//	DLGITEMTEMPLATE ̍\z
	CtrlListItem::CtrlTemplateArgs cta;
	cta.m_hdr = lpwExData;
	cta.m_id = (hbId << 8) + 0x10;	//	hbId  DlgPage ʂhc
	m_pCtrlList->initSequentialGet();
	CtrlListItem* cli;
	while ((cli = m_pCtrlList->getNextItem()) != NULL) {
		if (cli->getCtrlNum() > 0)	//	NewFormatCtrl 
			cli->createCtrlTemplate(cta);
	}

	//	_CAO̍쐬
	HWND hwndFrame = m_pDlgFrame->getUserDlg();
	m_bInTabCtrl = bInTabCtrl;
	m_hwndPage = ::CreateDialogIndirectParam(
					m_pDlgFrame->getSessionInstance()->getInstanceHandle(),
					(LPDLGTEMPLATE)(BYTE*)pDlgTemplate,
					hwndFrame,
					(DLGPROC)DlgPageProc,
					(LPARAM)this);
	if (m_hwndPage != NULL && hwndCtrl != NULL) {
		//	"root" ȊO̎q_CAÖʒu̕ύX
		RECT rct;
		::GetWindowRect(hwndCtrl,&rct);
		::ScreenToClient(hwndFrame,(LPPOINT)&rct);
		DWORD dlgunit = GetDialogBaseUnits(hwndFrame);
		::SetWindowPos(m_hwndPage,
						NULL,
						rct.left + ((x * LOWORD(dlgunit)) >> 2),
						rct.top + ((y * HIWORD(dlgunit)) >> 3),
						0, 0,
						SWP_NOSIZE|SWP_NOZORDER);
	}

	//	n
	m_dwPageSize = 0;

	if (m_pShowPageQueue->itemNum() > 0) {
		ShowPageQueueItem* childPageQI;
		while ((childPageQI = m_pShowPageQueue->getItemByIndex(0)) != NULL) {
			m_pDlgFrame->createPage(
								childPageQI->m_pdp->getName(),
								childPageQI->m_hwndCtrl,
								childPageQI->m_x,childPageQI->m_y,
								childPageQI->m_cx,childPageQI->m_cy,
								childPageQI->m_bInTabCtrl
							);
			m_pShowPageQueue->delItemByIndex(0);
		}
	}

	return m_hwndPage;
}

//	q_CAȌ
BOOL
DlgPage::initPage(HWND hDlg)
{
	if (hDlg == NULL) return FALSE;
	m_hwndPage = hDlg;
#if 0
	if (m_bInTabCtrl) {
		m_pDlgFrame->setBackGroundToTabColor(m_hwndPage);
	}
#else
	m_pDlgFrame->setBackGroundTheme(m_hwndPage, m_bInTabCtrl);
#endif
	//	Rg[̏
//	const StringBuffer& sbFocusedCtrl = m_pDlgFrame->getFocusedCtrlName();
	m_pCtrlList->initSequentialGet();
	CtrlListItem* cli;
	while ((cli = m_pCtrlList->getNextItem()) != NULL) {
		cli->onInitCtrl(hDlg);
		cli->enableCtrl(m_bEnable,FALSE);
#if 0
		if (sbFocusedCtrl.length() > 0 &&
			sbFocusedCtrl.compareTo(cli->getName()) == 0) {
			::SetFocus(cli->getCtrlHWND());
		}
#endif
	}
	::ShowWindow(m_hwndPage,m_bVisible ? SW_SHOW : SW_HIDE);
	return TRUE;
}

//	q_CAO̔j
BOOL
DlgPage::uninitPage()
{
	if (m_hwndPage == NULL) return FALSE;
	//	Rg[EBhȄI
	m_pCtrlList->initSequentialGet();
	CtrlListItem* cli;
	while ((cli = m_pCtrlList->getNextItem()) != NULL) {
		cli->onUninitCtrl();
	}
	m_hwndPage = NULL;
	return TRUE;
}

//	q_CAO
BOOL
DlgPage::showPage(BOOL bVisible)
{
	if (m_hwndPage == NULL) return FALSE;
	m_bVisible = bVisible;
	::ShowWindow(m_hwndPage,m_bVisible ? SW_SHOW : SW_HIDE);
	if (m_bVisible) {
		::SetWindowPos(m_hwndPage,HWND_TOP,0,0,0,0,
						SWP_NOMOVE|SWP_NOSIZE/*|SWP_SHOWWINDOW*/);
//		::InvalidateRect(m_hwndPage,NULL,TRUE);
		::RedrawWindow(m_hwndPage, NULL, NULL, RDW_ERASE | RDW_INVALIDATE | RDW_UPDATENOW);
	}

	//	Rg[̉
	m_pCtrlList->initSequentialGet();
	CtrlListItem* cli;
	while ((cli = m_pCtrlList->getNextItem()) != NULL) {
		cli->showCtrl(bVisible);
		cli->enableCtrl(cli->isEnabled(), FALSE);
	}

	return TRUE;
}

//	q_CAO̗L
BOOL
DlgPage::enablePage(BOOL bEnable)
{
	m_bEnable = bEnable;
	if (m_hwndPage != NULL) {
		m_pCtrlList->initSequentialGet();
		CtrlListItem* cli;
		while ((cli = m_pCtrlList->getNextItem()) != NULL) {
			cli->enableCtrl(bEnable,FALSE);
		}
	}
	return TRUE;
}

//	DLGITEMTEMPLATE ̃TCYƃEBhȆ傫̕]
//	LOWORD(return)	:	EBhE̕(P)
//	HIWORD(return)	:	EBhE̍(P)
DWORD
DlgPage::evalPageSize()
{
	if (m_pCtrlList == NULL) return 0;
	//	_CAȌI
	m_pCtrlList->addItem(new PageFormatCtrl("0",CTRLID_NEWPAGE));

	WORD	dcx = m_iniwidth, dcy = 0,
			py = 0, pcx = m_iniwidth, pcy = 0,
			cx = 0, ccx = m_iniwidth, ccy = 0,
			ly = 0, lcx = 0, lcy = 0,
			cw = 0, ch = 0;
	CtrlListItem* cli;
	m_cdit = 0;	//	ۂ̃Rg[̐
	m_dwPageSize = 0;	//	DLGITEMTEMPLATE ̃TCY
	m_pCtrlList->initSequentialGet();
	while ((cli = m_pCtrlList->getNextItem()) != NULL) {
		switch (cli->getCtrlType()) {
		case CTRLID_NEWCOLUMN:
			//	new column
			if (pcy < ccy) pcy = ccy;	//	y[W̒
			cx = pcx + 1;	//	Rʒu
			ccx = cli->getWidth();	//	R
			pcx += ccx + 1;	//	y[W̒
			ccy = ly = lcx = lcy = 0;
			break;
		case CTRLID_NEWPAGE:
			//	new page
			if (lcy > 0) ccy += lcy;	//	O̍s̍ǉ
			if (pcy < ccy) pcy = ccy;	//	y[W̒
			if (dcx < pcx) dcx = pcx;	//	_CAO̒
			dcy += pcy;	//	_CAO̒
			py = dcy;	//	y[Wcʒu
			pcx = ccx = cli->getWidth();	//	y[WER
			pcy = cx = ccy = ly = lcx = lcy = 0;
			break;
		case CTRLID_UNKNOWN:
			//	invalid data.
			m_pCtrlList->delItemByIndex(-1); // ǉ NewFormatCtrl 폜
			return 0L;
		default:
			m_cdit += cli->getCtrlNum();
			m_dwPageSize += (DWORD)cli->getDataSize();
			ch = cli->getHeight();
			if (ch < 1) ch = 3;	//	ListCtrl, ChkList, LViewCtrl ̏l
			cw = cli->getWidth();
			if (!cw || cw != 0xFFFF && cw > ccx - lcx) {
				//	wȂÃ݂RLw
				cw = ccx - lcx;
				cli->setRect(cx+lcx,py+ly,cw,ch);
				if (lcy < ch) lcy = ch;	//	s̒
				//	̍s
				ly += lcy;
				ccy = ly;
				lcx = 0;
				lcy = 0;
			} else {
				//	w肠
				if (cw == 0xFFFF) cw = (ccx - lcx)/2;	//	"half"
				cli->setRect(cx+lcx,py+ly,cw,ch);
				if (lcy < ch) lcy = ch;	//	s̒
				lcx += cw;	//	ʒu̒
				if (lcx > ccx) lcx = ccx;	//	R̒
			}
		}
	}
	//	end of defs.
	if (dcy == 0) dcy = 1;
	m_width = dcx;
	m_height = dcy;
	m_pCtrlList->delItemByIndex(-1); // ǉ NewFormatCtrl 폜
	return (DWORD)MAKELONG(m_width,m_height);
}

BOOL
DlgPage::addShowPageQueue(
	LPCSTR pname,
	HWND hwndCtrl,
	WORD x, WORD y, WORD cx, WORD cy,
	bool bInTabCtrl)
{
	if (m_pDlgFrame->getUserDlg() == NULL) return FALSE;
	DlgPage* pdp = m_pDlgFrame->getPage(pname);
	if (pdp == NULL) return FALSE;
	m_pShowPageQueue->addItem(new ShowPageQueueItem(pdp, hwndCtrl,
													x, y, cx, cy,
													bInTabCtrl),
							  -1);
	return TRUE;
}

WORD
DlgPage::getDefID() const
{
	CtrlListItem* cli;
	m_pCtrlList->initSequentialGet();
	while ((cli = m_pCtrlList->getNextItem()) != NULL) {
		WORD id = cli->getDefID();
		if (id != 0xFFFF) return id;
	}
	return 0xFFFF;
}

//	Rg[̒ǉ
int
DlgPage::addCtrl(CtrlListItem* cli)
{
	if (cli == NULL) return 0;
	cli->setParentPage(this);	//	e_CAÕZbg
	const StringBuffer& name = cli->getName();
	if (name.length() > 0) // Rg[̖O̓o^
		m_pHashCtrl->setValue(name,cli);
	m_pCurCtrlData = cli;	//	Poke ̑Ώ
	return m_pCtrlList->addItem(cli);
}

int
DlgPage::addCtrl(int type, const StringBuffer& name, const StringBuffer& text)
{
	return this->addCtrl(CtrlListItem::createCtrl(type,name,text));
}

//	Rg[̎擾
CtrlListItem*
DlgPage::getCtrl(const StringBuffer& name)
{
	return (name.length() > 0) ? m_pHashCtrl->getValue(name) : m_pCurCtrlData;
}

CtrlListItem*
DlgPage::prevCtrl(const CtrlListItem* pCtrl)
{
	CtrlListItem*	cli;
	CtrlListItem*	pPrevCtrl = NULL;

	m_pCtrlList->initSequentialGet();
	while ((cli = m_pCtrlList->getNextItem()) != NULL)
	{
		if (cli == pCtrl)
		{
			return pPrevCtrl;
		}

		pPrevCtrl = cli;
	}

	return NULL;
}

CtrlListItem*
DlgPage::nextCtrl(const CtrlListItem* pCtrl)
{
	CtrlListItem*	cli;

	m_pCtrlList->initSequentialGet();
	while ((cli = m_pCtrlList->getNextItem()) != NULL)
	{
		if (cli == pCtrl)
		{
			return m_pCtrlList->getNextItem();
		}
	}

	return NULL;
}

HWND
DlgPage::getFocusedCtrl()
{
	if (m_hwndPage == NULL) return NULL;
	CtrlListItem* cli;
	HWND hwndCtrl;
	m_pCtrlList->initSequentialGet();
	while ((cli = m_pCtrlList->getNextItem()) != NULL) {
		if ((hwndCtrl = cli->getFocusedCtrl()) != NULL) return hwndCtrl;
	}
	return NULL;
}

static CtrlListItem*
GetCtrlListItemByHWND(HWND hwndCtrl)
{
	if (hwndCtrl == NULL) return NULL;
	CtrlListItem*
		cli = (CtrlListItem*)::SendMessage(hwndCtrl, WM_GET_CTRL_PTR, 0, 0);
	if (cli == NULL || !cli->isValid()) {
		return NULL;
	}
	return cli;
}

static inline CtrlListItem*
GetCtrlListItemByID(HWND hwndPage, WORD id)
{
	if (hwndPage == NULL) return NULL;
	return GetCtrlListItemByHWND(::GetDlgItem(hwndPage, id));
}

//	WM_COMMAND nh
WORD
DlgPage::onCommand(WPARAM wParam, LPARAM lParam)
{
	CtrlListItem* cli = GetCtrlListItemByID(m_hwndPage, LOWORD(wParam));
	return cli ? cli->onWmCommand(wParam, lParam) : 0xFFFF;
}

//	WM_NOTIFY nh
WORD
DlgPage::onNotify(WPARAM wParam, LPARAM lParam)
{
	NMHDR* pnmh = reinterpret_cast<NMHDR*>(lParam);
	CtrlListItem* cli = GetCtrlListItemByID(m_hwndPage, pnmh->idFrom);
	return cli ? cli->onWmNotify(wParam, lParam) : 0xFFFF;
}

//	WM_HSCROLL nh(gbNo[̂߂ɂ̂ݑ݂)
WORD
DlgPage::onHScroll(WPARAM wParam, LPARAM lParam)
{
	CtrlListItem* cli = GetCtrlListItemByHWND((HWND)lParam);
	return cli ? cli->onWmCommand(wParam, lParam) : 0xFFFF;
}

//	WM_CTLCOLOR nh
HBRUSH
DlgPage::onCtlColor(UINT uMsg, WPARAM wParam, LPARAM lParam)
{
	CtrlListItem* cli = GetCtrlListItemByHWND((HWND)lParam);
	return cli ? cli->onWmCtlColor(uMsg, wParam, lParam) : NULL;
}

WORD
DlgPage::onImeNotify(WPARAM wParam, LPARAM lParam)
{
	CtrlListItem* cli = GetCtrlListItemByHWND((HWND)lParam);
	return cli ? cli->onWmImeNotify(wParam, lParam) : 0xFFFF;
}

//	_CAOf[^̏
BOOL
DlgPage::dumpData(DlgDataFile& ddfile)
{
	if (!ddfile.isValid()) return FALSE;
	StringBuffer secname(GetString(STR_DLGDATA_DLGPAGEPROPERTY),-1,6);
	secname.append((TCHAR)'_').append(m_strName);
	ddfile.write(m_strName,GetString(STR_DLGDATA_PAGENAME),secname);
	ddfile.write(m_bEnable,GetString(STR_DLGDATA_ENABLE),secname);
	ddfile.write(m_iniwidth,GetString(STR_DLGDATA_WIDTH),secname);
	int	num = m_pCtrlList->itemNum();
	ddfile.write(num,GetString(STR_DLGDATA_CTRLNUM),secname);
	secname.reset(GetString(STR_DLGDATA_CTRLPROPERTY));
	secname.append((TCHAR)'_').append(m_strName).append((TCHAR)':');
	int	len = secname.length();
	m_pCtrlList->initSequentialGet();
	CtrlListItem* cli;
	for (int i = 0; i < num; i++) {
		secname.setlength(len);
		secname.append(i);
		ddfile.setSecName(secname);
#ifdef _DEBUG
		char buf[80];
		wsprintf(buf, "dump:%s\n", (LPCSTR)secname);
		::OutputDebugString(buf);
#endif
		cli = m_pCtrlList->getNextItem();
		cli->dumpData(ddfile);
	}
	return TRUE;
}

//	_CAOf[^̓Ǎ
BOOL
DlgPage::loadData(DlgDataFile& ddfile)
{
	if (!ddfile.isValid()) return FALSE;
	StringBuffer secname(GetString(STR_DLGDATA_DLGPAGEPROPERTY),-1,6);
	secname.append((TCHAR)'_').append(m_strName);
	ddfile.read(&m_bEnable,GetString(STR_DLGDATA_ENABLE),secname);
	ddfile.read(&m_iniwidth,GetString(STR_DLGDATA_WIDTH),secname);
	int	num;
	ddfile.read(&num,GetString(STR_DLGDATA_CTRLNUM),secname);
	secname.reset(GetString(STR_DLGDATA_CTRLPROPERTY));
	secname.append((TCHAR)'_').append(m_strName).append((TCHAR)':');
	int	len = secname.length();
	StringBuffer namebuf(32), textbuf(80);
	for (int i = 0; i < num; i++) {
		secname.setlength(len);
		secname.append(i);
		int type;
		ddfile.read(&type, GetString(STR_DLGDATA_TYPE), secname);
		ddfile.read(namebuf, GetString(STR_DLGDATA_NAME), secname);
		ddfile.read(textbuf, GetString(STR_DLGDATA_TEXT), secname);
		m_pCurCtrlData = CtrlListItem::createCtrl(type,namebuf,textbuf);
		if (m_pCurCtrlData == NULL) continue;
		ddfile.setSecName(secname);
#ifdef _DEBUG
		char buf[80];
		wsprintf(buf, "load:%s\n", (LPCSTR)secname);
		::OutputDebugString(buf);
#endif
		if (m_pCurCtrlData->loadData(ddfile)) this->addCtrl(m_pCurCtrlData);
		else {
			delete m_pCurCtrlData;
			m_pCurCtrlData = NULL;
		}
	}
	return TRUE;
}

int
DlgPage::setExProperty(const StringBuffer& key, const StringBuffer& value)
{
	DEBUG_OUTPUT(("Enter DlgPage(%s)::setExProperty(%s, %s)", (LPCSTR)m_strName, (LPCSTR)key, (LPCSTR)value));

	if (key.compareTo("bgcolor") == 0)
	{
		if (value.length() == 0)
		{
			m_crefBkColor = COLOR_3DFACE;
			if (m_hbkBrush != NULL)
			{
				::DeleteObject(m_hbkBrush);
				m_hbkBrush = NULL;
			}
		}
		else
		{
			LOGBRUSH	logBrush;
			memset(&logBrush, 0, sizeof(logBrush));
			logBrush.lbColor = ColorTable::colorStrToColorRef(value);
			logBrush.lbStyle = BS_SOLID;

			HBRUSH	hbkBrush = ::CreateBrushIndirect(&logBrush);
			if (hbkBrush == NULL)
			{
				DEBUG_OUTPUT(("Failed to create brush in DlgPage(%s)::setExProperty(%s, %s)", (LPCSTR)m_strName, (LPCSTR)key, (LPCSTR)value));
				return 0;
			}

			m_crefBkColor = logBrush.lbColor;
			if (m_hbkBrush != NULL)
			{
				::DeleteObject(m_hbkBrush);
			}
			m_hbkBrush = hbkBrush;
		}
	}

	int	ret;

	if (m_strName.compareTo("root") != 0)
	{
		if (m_hwndPage != NULL)
		{
//			::InvalidateRect(m_hwndPage, NULL, TRUE);
//			::UpdateWindow(m_hwndPage);
			::RedrawWindow(m_hwndPage, NULL, NULL, RDW_ERASE | RDW_INVALIDATE | RDW_ALLCHILDREN | RDW_UPDATENOW);
		}
		ret = 1;
	}
	else
	{
		ret = m_pDlgFrame->setExProperty(key, value);
	}

	DEBUG_OUTPUT(("Leave DlgPage(%s)::setExProperty(%s, %s)", (LPCSTR)m_strName, (LPCSTR)key, (LPCSTR)value));

	return ret;
}

StringBuffer
DlgPage::getExProperty(const StringBuffer& key)
{
	if (key.compareTo("bgcolor") == 0)
	{
		if (m_hbkBrush != NULL)
		{
			return ColorTable::colorRefToColorStr(m_crefBkColor);
		}
		else
		{
			return ColorTable::colorRefToColorStr(::GetSysColor(COLOR_3DFACE));
		}
	}

	return nullStr;
}

//	e_CAÕR[obN֐
BOOL CALLBACK
DlgFrameProc(HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
	DlgFrame* pdf;

	if (uMsg == WM_INITDIALOG) {
		//	e_CAȌ
		pdf = reinterpret_cast<DlgFrame*>(lParam);
		if (!pdf->initFrame(hDlg)) {
			::DestroyWindow(hDlg);
		}
		::SetWindowLongPtr(hDlg, DWLP_USER, (LONG_PTR)lParam);
		return FALSE;
	}

	if ((pdf = reinterpret_cast<DlgFrame*>(::GetWindowLongPtr(hDlg,DWLP_USER)))
		== NULL) return FALSE;

	switch (uMsg) {
	case WM_USER_MESSAGEBOX:
		if (wParam) {
			//	bZ[W̕\
			StringBuffer* msg = reinterpret_cast<StringBuffer*>(wParam);
			if (::MessageBox(hDlg,
							*msg,
							GetString(STR_USRMSGBOX_MESSAGE),
							MB_OK|MB_ICONINFORMATION) == IDOK) {
				StringBuffer buf(32);
				pdf->setNotify(buf.append((ULONG)lParam));
			}
			delete msg;
		}
		break;

	case WM_USER_QUESTIONBOX:
		if (wParam) {
			//	mFbZ[W̕\
			StringBuffer* msg = reinterpret_cast<StringBuffer*>(wParam);
			StringBuffer buf(32);
			if (::MessageBox(hDlg,
							*msg,
							GetString(STR_USRMSGBOX_QUESTION),
							MB_YESNO|MB_ICONQUESTION) == IDYES) {
				buf.append((int)HIWORD(lParam));
			} else {
				buf.append((int)LOWORD(lParam));
			}
			pdf->setNotify(buf);
			delete msg;
		}
		break;

	case WM_USER_NOTIFY:
		{
			//	q_CAO̒ʒmR[h
			DEBUG_OUTPUT(("WM_USER_NOTIFY handled"));
			pdf->setNotify(StringBuffer(16).append((int)wParam));
		}
		break;

	case WM_USER_GETFOCUSEDCTRL:
		{
			::SetWindowLongPtr(hDlg, DWLP_MSGRESULT, (LONG_PTR)(LPCSTR)pdf->getFocusedCtrl());
		}
		break;

	case WM_USER_SETFOCUSEDCTRL:
		pdf->setFocusedCtrl();
		break;

	case DM_SETDEFID:
		// prevent the default message handling
		break;

	case DM_GETDEFID:
		{
			WORD wDefID = pdf->getDefID();
#if 1
			if (wDefID == 0xFFFF) {
				// ftHgvbV{^݂Ȃ
				wDefID = IDOK;
			}
#endif
			::SetWindowLongPtr(hDlg, DWLP_MSGRESULT, (DC_HASDEFID << 16) | wDefID);
		}
		break;

	case WM_CTLCOLORDLG:
		{
			HBRUSH	hBrush = pdf->getBkBrush();
			DEBUG_OUTPUT(("WM_CTLCOLORDLG is posted to DlgFrame, returns 0x%08x", (DWORD)hBrush));
			return (BOOL)hBrush;
		}

	case WM_COMMAND:
		{
//			::OutputDebugString("Enter DlgFrame WM_COMMAND(1)\n");
			//	^[L[ADlgFrame  DM_GETDEFID A
			//	̌㑗Ă WM_COMMAND 𐳂 DlgPage ɓn
			int	index = HIBYTE(LOWORD(wParam));	//	ctrlID  HIBYTE 
												//	DlgPage  index
			if (index > 0) {
//				::OutputDebugString("Enter DlgFrame WM_COMMAND(2)\n");
				DlgPage* pdp = pdf->getPage(--index);
				if (pdp != NULL) {
					::PostMessage(pdp->gethwndPage(),WM_COMMAND,wParam,lParam);
					return TRUE;
				}
			} else if (LOWORD(wParam) == IDOK || LOWORD(wParam) == IDCANCEL) {
//				::OutputDebugString("Enter DlgFrame WM_COMMAND(3)\n");
				return ::PostMessage(hDlg,
									WM_USER_NOTIFY,
									(LOWORD(wParam) == IDOK),
									0);
			}
		}
		return FALSE;

	case WM_SYSCOMMAND:
		if (wParam != SC_CLOSE) return FALSE;
//		::OutputDebugString("Enter WM_SYSCOMMAND in DlgFrame\n");
		::SendMessage(hDlg,WM_USER_NOTIFY,0,0);	//	"0" ʒm
		//	pdf->setDialogClosed()Ă΂Ȃ!!
		break;

	case WM_CLOSE:
//		::OutputDebugString("Enter WM_CLOSE in DlgFrame\n");
//		::SendMessage(hDlg,WM_USER_NOTIFY,0,0);	//	"0" ʒm
		::SendMessage(hDlg,WM_USER_NOTIFY,0,0);	//	"0" ʒm
		pdf->setDialogClosed();
		DEBUG_OUTPUT(("WM_CLOSE handled"));
//		::DestroyWindow(hDlg);
		return FALSE;

	case WM_DESTROY:
		{
			HWND hwndOwner = ::GetParent(hDlg);
			if (hwndOwner != NULL) {
//	[hX_CAOƊ؂āAȉ̃R[h͂΂(^^;폜B
//	GۃEBhENbNƁu}NI܂HvƏo邯
//	܂ȂB
//				::EnableWindow(hwndOwner,TRUE);
				::SetForegroundWindow(hwndOwner);
			}
			pdf->uninitFrame();
			::PostQuitMessage(0L);
		}
		break;

	default:
		return FALSE;
	}
	return TRUE;
}

DlgFrame::DlgFrame()
	:	m_pSessionInstance(NULL),
		m_hwndFrame(NULL),
		m_pThemeWrapper(NULL),
		m_pPageList(new NamedLinkList<DlgPage>),
		m_pCurDlgPage(NULL),
		m_DlgTitle(80),
		m_pFontProp(new FontProperty),
		m_width(0),
		m_height(0),
		m_flags(0),
		m_imestate(0),
		m_bAlreadyFocused(FALSE),
		m_crefBkColor(COLOR_3DFACE),
		m_hbkBrush(NULL)
{
	try {
		m_pThemeWrapper = new ThemeWrapper();
		m_pThemeWrapper->SetThemeAppProperties(STAP_ALLOW_NONCLIENT |
											   STAP_ALLOW_CONTROLS  |
											   STAP_ALLOW_WEBCONTENT);
	} catch (ThemeNotSupportedError&) {
		m_pThemeWrapper = NULL;
	}

	GetDefFontProperty(m_pFontProp);
	m_pos.x = m_pos.y = 0;
}

DlgFrame::~DlgFrame()
{
	if (m_hwndFrame != NULL) ::DestroyWindow(m_hwndFrame);
	delete m_pPageList;
	delete m_pThemeWrapper;
	delete m_pFontProp;
}

void
DlgFrame::setNotify(const StringBuffer& newvalue)
{
	if (!m_bAlreadyClosed) {
		m_pSessionInstance->setNotify(newvalue);
	}
}

BOOL
DlgFrame::getNotify(StringBuffer& buf, DWORD wait) const
{
	return m_pSessionInstance->getNotify(buf,wait);
}

void
DlgFrame::resetNotify()
{
	m_pSessionInstance->resetNotify();
}

void
DlgFrame::setFrameProperty(
	const StringBuffer& title,
	POINTS* ppt,
	WORD wflags,
	const StringBuffer& fontname,
	WORD fontsize,
	WORD wPosUnit)
{
	m_DlgTitle = title;
	if (ppt != NULL) m_pos = *ppt;
	m_flags = wflags;
	if (fontname.length() > 0) m_pFontProp->m_fontname = fontname;
	if (fontsize > 0) m_pFontProp->m_fontsize = fontsize;
	m_wPosUnit = wPosUnit;
}

//	e_CAO̐
HWND
DlgFrame::createFrame(HWND hwndOwner, BOOL bOnTop)
{
	//	"root" q_CAO̓o^̊mF
	m_pCurDlgPage = this->getPage(strRootPageName);
	if (m_pCurDlgPage == NULL) return NULL;

	//	"root" q_CAȎ傫vZ
	DWORD size = m_pCurDlgPage->evalPageSize();
	m_width = LOWORD(size) + 2;
	m_height = HIWORD(size) + 1;

	//	e_CAO DLGTEMPLATE ̐
	int	bufsize = 2 + sizeof(DLGTEMPLATE) + 
//				sizeof(WORD) * (m_DlgTitle.length() + 
//								m_pFontProp->m_fontname.length() + 2) + 
				((m_DlgTitle.length() + 
				 m_pFontProp->m_fontname.length() + 2) << 1) + 
				16;
	Array<BYTE> pDlgTemplate(bufsize);
	pDlgTemplate.zero();

	CreateDialogTemplate((LPDLGTEMPLATE)(BYTE*)pDlgTemplate,
						WS_POPUP|WS_CAPTION|WS_SYSMENU|
							DS_MODALFRAME|DS_SETFONT,
						0L,
						0,
						m_pos.x*4, m_pos.y*8,
						m_width*UWIDTH, m_height*UHEIGHT,
						m_DlgTitle,
						m_pFontProp->m_fontname,
						(WORD)m_pFontProp->m_fontsize);
	if (bOnTop) m_flags |= 0x10;	//	alwaysontop

	m_bAlreadyClosed = FALSE;

	//	e_CAÕEBhE̐
	m_hwndFrame = ::CreateDialogIndirectParam(
								m_pSessionInstance->getInstanceHandle(),
								(LPDLGTEMPLATE)(BYTE*)pDlgTemplate,
								hwndOwner,
								(DLGPROC)DlgFrameProc,
								(LPARAM)this);

	return m_hwndFrame;
}

//	_CAO̕\
BOOL
DlgFrame::showFrame()
{
	if (m_hwndFrame == NULL) return FALSE;

	//	_CAÖʒu̕ύX
	POINTS pos = GetDlgPosByPixel(m_pos, m_wPosUnit, m_hwndFrame);
	POINTS org = GetDlgPosOrigin(m_flags & 0x000F, m_hwndFrame);

	switch (m_flags & 0x000F) {
	case DLGPOS_SCREEN_COORD:
	case DLGPOS_CLIENT_COORD:
	case DLGPOS_CARET_COORD:
	case DLGPOS_CURSOR_COORD:
		pos.x += org.x;
		pos.y += org.y;
		break;
	case DLGPOS_SCREEN_CENTER:
	case DLGPOS_CLIENT_CENTER:
	default:
		pos = org;
		break;
	}

#if 0
	RECT rect, drect;
	::GetWindowRect(m_hwndFrame, &rect);

	HWND hwndParent = ::GetParent(m_hwndFrame);

	switch (m_flags & 0x000F) {
	case DLGPOS_SCREEN_COORD:
		/* Screen Coordinate */
		break;
	case DLGPOS_CLIENT_COORD:
		/*	Window Coordinate	*/
		::ClientToScreen(hwndParent, &pos);
		break;
	case DLGPOS_SCREEN_CENTER:
		/*	Screen Center	*/
		::GetWindowRect(::GetDesktopWindow(), &drect);
		pos.x = ((drect.left + drect.right) >> 1)
				- ((rect.right - rect.left) >> 1);
		pos.y = ((drect.top + drect.bottom) >> 1)
				- ((rect.bottom - rect.top) >> 1);
		break;
	case DLGPOS_CARET_COORD:
		/*  Caret Coordinate  */
		if (::GetCaretPos(&cpos)) {
			::ClientToScreen(hwndParent, &cpos);
			pos.x += cpos.x;
			pos.y += cpos.y;
		}
		break;
	case DLGPOS_CURSOR_COORD:
		/*  Cursor Coordinate  */
		if (::GetCursorPos(&cpos)) {
			pos.x += cpos.x;
			pos.y += cpos.y;
		}
		break;
	case DLGPOS_CLIENT_CENTER:
	default:
		::GetWindowRect(hwndParent, &drect);
		pos.x = ((drect.left + drect.right) >> 1)
				- ((rect.right - rect.left) >> 1);
		pos.y = ((drect.top + drect.bottom) >> 1)
				- ((rect.bottom - rect.top) >> 1);
	}
#endif

	HWND hwndIA = HWND_TOP;
	UINT uFlag = SWP_NOSIZE;
	if ((m_flags & 0x10) != 0) hwndIA = HWND_TOPMOST;	//	alwaysontop
	::SetWindowPos(m_hwndFrame, hwndIA, pos.x, pos.y, 0, 0, uFlag);

//	[hX_CAOƊ؂āAȉ̃R[h͂΂(^^;폜B
//	GۃEBhENbNƁu}NI܂HvƏo邯
//	܂ȂB
//	HWND hwndOwner = ::GetParent(m_hwndFrame);
//	if (hwndOwner != NULL) ::EnableWindow(hwndOwner,FALSE);

	// ftHgtH[JX̐ݒ
//	this->setFocusedCtrl(m_sbFocusedCtrl);
	DlgPage* pdproot = this->getPage(strRootPageName);
	if (pdproot != NULL) {
		pdproot->showPage(TRUE);
	}

	// ftHgtH[JX̐ݒ
//	this->setFocusedCtrl(m_sbFocusedCtrl);

	::ShowWindow(m_hwndFrame, SW_SHOW);

	return TRUE;
}

//	e_CAO̔j
BOOL
DlgFrame::closeFrame()
{
	if (m_hwndFrame == NULL) return FALSE;
	BOOL ret = !::SendMessage(m_hwndFrame, WM_DESTROY, 0, 0);
	m_hwndFrame = NULL;
	return ret;
}

//	e_CAOf[^̍ď
void
DlgFrame::resetFrame()
{
	if (m_hwndFrame != NULL) return;
	m_pCurDlgPage = NULL;
	m_DlgTitle.reset();
	GetDefFontProperty(m_pFontProp);
	m_pos.x = m_pos.y = 0;
	m_width = 0;
	m_flags = 0;
}

//	e_CAȌ
BOOL
DlgFrame::initFrame(HWND hDlg)
{
	if (hDlg == NULL) return FALSE;

	m_hwndFrame = hDlg;

	//	"root" q_CAO̐
	HWND hwndRoot = this->createPage(strRootPageName,
									 NULL,
									 UWIDTH,UHEIGHT / 2,
									 m_width-2,m_height - 1,
									 false);

#if 0
	if (hwndRoot) {
		// eXbhɒʒm
		// (̃^C~OōsȂ SetFocus() ŃnO)
		m_pSessionInstance->SessionInstance::setNotify(okStr);
		// ftHgtH[JX̐ݒ
		this->setFocusedCtrl(m_sbFocusedCtrl);
	} else {
		m_pSessionInstance->SessionInstance::setNotify(ngStr);
	}
#endif
	m_bAlreadyFocused = FALSE;

	return hwndRoot != NULL;
}

//	e_CAO̔j
BOOL
DlgFrame::uninitFrame()
{
	m_hwndFrame = NULL;
	if (m_pPageList != NULL) {
		DlgPage* pDlgPage;
		m_pPageList->initSequentialGet();
		while ((pDlgPage = m_pPageList->getNextItem()) != NULL)
			pDlgPage->uninitPage();
	}
	return TRUE;
}

//	q_CAO̒ǉ
int
DlgFrame::addPage(const StringBuffer& name, WORD width)
{
	if (name.length() <= 0 || m_hwndFrame != NULL) return 0;
	m_pCurDlgPage = new DlgPage(this,name,width);	//	Poke ̑Ώ
	return m_pPageList->addItem(m_pCurDlgPage);
}

//	Rg[ǉΏۂ̎q_CAO̕ύX
BOOL
DlgFrame::setCurrentPage(const StringBuffer& name)
{
	if (name.length() <= 0 || m_hwndFrame != NULL) return FALSE;
	DlgPage* pdp = this->getPage(name);
	if (pdp == NULL) return FALSE;
	m_pCurDlgPage = pdp;
	return TRUE;
}

//	q_CAOEBhE̐
HWND
DlgFrame::createPage(
	const StringBuffer& name,
	HWND hWnd,
	WORD ofx, WORD ofy,
	WORD cx, WORD cy,
	bool bInTabCtrl)
{
	//	ofx, ofy ̓_CAOP(=/4, c=/8)A
	//	cx, cy ͕Pʂŗ^
	if (m_hwndFrame == NULL) return NULL;
	DlgPage* pdp = this->getPage(name);
	if (pdp == NULL) return NULL;
	//	Rg[ID HIBYTE() ɂ͎q_CAOindex
	BYTE index = (BYTE)(m_pPageList->getItemIndexByPtr(pdp) + 1);
	return pdp->createPage(hWnd, index,
						   ofx, ofy, cx*UWIDTH, cy*UHEIGHT,
						   bInTabCtrl, *m_pFontProp);
}

//	Rg[𖼑OŎ擾
CtrlListItem*
DlgFrame::getCtrl(const StringBuffer& ctrlname, const StringBuffer& pagename) const
{
	if (ctrlname.length() <= 0 && pagename.length() <= 0) {
		//	݂̎q_CAO猻݂̃Rg[擾
		if (m_pCurDlgPage == NULL) return NULL;
		return m_pCurDlgPage->getCtrl();
	} else if (ctrlname.length() > 0) {
		//	w肳ꂽRg[SĂ̎q_CAO猟
		DlgPage* pdp;
		CtrlListItem* cli = NULL;
		m_pPageList->initSequentialGet();
		while ((pdp = m_pPageList->getNextItem()) != NULL) {
			if ((cli = pdp->getCtrl(ctrlname)) != NULL) break;
		}
		return cli;
	} else {
		//	wq_CAOwRg[擾
		DlgPage* pdp = m_pPageList->getItemByName(pagename);
		if (pdp == NULL) return NULL;
		return pdp->getCtrl(ctrlname);
	}
	//	not reached.
}

int
DlgFrame::setFocusedCtrl(const StringBuffer& name)
{
	if (name.length() <= 0) {
		m_sbFocusedCtrl = nullStr;
	} else {
		CtrlListItem* pctrl = this->getCtrl(name);
		if (pctrl == NULL) {
			return FALSE;
		}
		m_sbFocusedCtrl = name;
	}

	if (m_hwndFrame != NULL) {
		::PostMessage(m_hwndFrame, WM_USER_SETFOCUSEDCTRL, 0, 0);
	}

	return TRUE;
}

int
DlgFrame::setFocusedCtrl()
{
	if (m_sbFocusedCtrl.length() <= 0) {
		DlgPage* pdproot = this->getPage(strRootPageName);
		if (pdproot == NULL) return FALSE;
		HWND hwndFocused = pdproot->getFocusedCtrl();
		if (hwndFocused) {
			::SetFocusForced(hwndFocused);
		}
	} else {
		CtrlListItem* pctrl = this->getCtrl(m_sbFocusedCtrl);
		if (pctrl == NULL) {
			return FALSE;
		}

		HWND hwndFocused = pctrl->getCtrlHWND();
		if (hwndFocused == NULL) {
			return FALSE;
		}

		::SetFocusForced(hwndFocused);
	}
	return TRUE;
}

const StringBuffer&
DlgFrame::getFocusedCtrl() const
{
	if (m_hwndFrame == NULL) return m_sbFocusedCtrl;
	HWND hCtrl = ::GetFocus();
	if (hCtrl != NULL) {
		CtrlListItem*
			pCtrl = (CtrlListItem*)::SendMessage(hCtrl, WM_GET_CTRL_PTR, 0, 0);
		if (pCtrl != NULL && pCtrl->isValid()) {
			return pCtrl->getName();
		}
	}
	return nullStr;
}

int
DlgFrame::setCtrlFocusedItem(const StringBuffer& ctrl, const StringBuffer& item)
{
	CtrlListItem* pCtrl = getCtrl(ctrl);
	if (!pCtrl) {
		return 0;
	}

	return pCtrl->onSetFocusedItem(item);
}

StringBuffer
DlgFrame::getCtrlFocusedItem(const StringBuffer& ctrl) const
{
	CtrlListItem* pCtrl = getCtrl(ctrl);
	if (!pCtrl) {
		return nullStr;
	}

	return pCtrl->onGetFocusedItem();
}

int
DlgFrame::setImeState(int nState)
{
	m_imestate = nState;
	m_bAlreadyFocused = FALSE;
#if 0 // ̓tH[JX𓾂Rg[Ɍʂɐݒ肪Kv
	if (m_hwndFrame) {
		HIMC hImc = ::ImmGetContext(m_hwndFrame);
		switch (m_imestate) {
		case 1:
			if (!::ImmGetOpenStatus(hImc))
				::ImmSetOpenStatus(hImc, TRUE);
			break;
		default:
			if (::ImmGetOpenStatus(hImc))
				::ImmSetOpenStatus(hImc, FALSE);
			break;
		}
		::ImmReleaseContext(m_hwndFrame, hImc);
	}
#endif
	return m_imestate;
}

WORD
DlgFrame::getDefID() const
{
	if (m_hwndFrame == NULL) return 0xFFFF;

	DlgPage* pdpRoot = ((DlgFrame*)this)->getPage(strRootPageName);
	if (pdpRoot == NULL) return 0xFFFF;

	return pdpRoot->getDefID();
}

POINTS
DlgFrame::getDlgPos() const
{
	if (!m_hwndFrame) {
		return m_pos;
	}

	RECT rcFrame;
	::GetWindowRect(m_hwndFrame, &rcFrame);

	POINTS pos;
	pos.x = (WORD)rcFrame.left;  pos.y = (WORD)rcFrame.top;

	POINTS org = GetDlgPosOrigin(m_flags & 0x000F, m_hwndFrame);
	pos.x -= org.x;  pos.y -= org.y;

	pos = GetDlgPosByUnit(pos, m_wPosUnit, m_hwndFrame);

	return pos;
}

void
DlgFrame::setDlgPos(const POINTS& pos, WORD wOrigin, WORD wUnit)
{
	m_pos = pos;
	m_wPosUnit = wUnit;
	m_flags = (m_flags & 0xF0) | wOrigin;

	if (m_hwndFrame) {
		POINTS newpos = GetDlgPosByPixel(m_pos, m_wPosUnit, m_hwndFrame);
		POINTS org = GetDlgPosOrigin(m_flags & 0xF, m_hwndFrame);
		newpos.x += org.x;  newpos.y += org.y;
		HWND hwndIA = HWND_TOP;
		UINT uFlag = SWP_NOSIZE;
		if ((m_flags & 0x10) != 0) hwndIA = HWND_TOPMOST;	//	alwaysontop
		::SetWindowPos(m_hwndFrame, hwndIA, newpos.x, newpos.y, 0, 0, uFlag);
	}
}

//	_CAOf[^̏
BOOL
DlgFrame::dumpData(DlgDataFile& ddfile)
{
	if (!ddfile.isValid()) return FALSE;
	LPCSTR	secname = GetString(STR_DLGDATA_DLGFRAMEPROPERTY);
	ddfile.write(m_DlgTitle,GetString(STR_DLGDATA_TITLE),secname);
	ddfile.write(m_pFontProp->m_fontsize,
				GetString(STR_DLGDATA_FONTSIZE),secname);
	ddfile.write(m_pFontProp->m_fontname,
				GetString(STR_DLGDATA_FONTNAME),secname);
	ddfile.write(m_pos.x,GetString(STR_DLGDATA_PLACEX),secname);
	ddfile.write(m_pos.y,GetString(STR_DLGDATA_PLACEY),secname);
	ddfile.write(m_width,GetString(STR_DLGDATA_WIDTH),secname);
	ddfile.write(m_flags,GetString(STR_DLGDATA_FLAGS),secname);
	m_pPageList->initSequentialGet();
	DlgPage* pdp;
	StringBuffer buf(GetString(STR_DLGDATA_PAGENAME),-1,6);
	buf.append((TCHAR)'_');
	int	len = buf.length(), num = m_pPageList->itemNum();
	for (int i = 0; i < num; i++) {
		pdp = m_pPageList->getNextItem();
		buf.setlength(len);
		buf.append(i);
		ddfile.write(pdp->getName(),buf,secname);
		pdp->dumpData(ddfile);
	}
	ddfile.write(num,GetString(STR_DLGDATA_PAGENUM),secname);
	return TRUE;
}

//	_CAOf[^̓Ǎ
BOOL
DlgFrame::loadData(DlgDataFile& ddfile)
{
	if (!ddfile.isValid()) return FALSE;
	LPCSTR secname = GetString(STR_DLGDATA_DLGFRAMEPROPERTY);
	ddfile.read(m_DlgTitle,GetString(STR_DLGDATA_TITLE),secname);
	ddfile.read(&m_pFontProp->m_fontsize,
				GetString(STR_DLGDATA_FONTSIZE),secname);
	ddfile.read(m_pFontProp->m_fontname,
				GetString(STR_DLGDATA_FONTNAME),secname);
	ddfile.read((WORD*)&m_pos.x,GetString(STR_DLGDATA_PLACEX),secname);
	ddfile.read((WORD*)&m_pos.y,GetString(STR_DLGDATA_PLACEY),secname);
	ddfile.read(&m_width,GetString(STR_DLGDATA_WIDTH),secname);
	ddfile.read(&m_flags,GetString(STR_DLGDATA_FLAGS),secname);
	int	num;
	ddfile.read(&num,GetString(STR_DLGDATA_PAGENUM),secname);
	DlgPage* pdp;
	StringBuffer keybuf(GetString(STR_DLGDATA_PAGENAME),-1,6);
	keybuf.append((TCHAR)'_');
	int	len = keybuf.length();
	StringBuffer buf(80);
	for (int i = 0; i < num; i++) {
		keybuf.setlength(len);
		keybuf.append(i);
		ddfile.read(buf,keybuf,secname);
		if (buf.length() <= 0) continue;
		this->addPage(buf,1);
		pdp = this->getPage(buf);
		pdp->loadData(ddfile);
	}
	return TRUE;
}

int
DlgFrame::setExProperty(const StringBuffer& key, const StringBuffer& value)
{
	DEBUG_OUTPUT(("Enter DlgFrame::setExProperty(%s, %s)", (LPCSTR)key, (LPCSTR)value));

	if (key.compareTo("bgcolor") != 0)
	{
		return 0;
	}

	if (value.length() == 0)
	{
		m_crefBkColor = COLOR_3DFACE;
		if (m_hbkBrush != NULL)
		{
			::DeleteObject(m_hbkBrush);
			m_hbkBrush = NULL;
		}
	}
	else
	{
		LOGBRUSH	logBrush;
		memset(&logBrush, 0, sizeof(logBrush));
		logBrush.lbColor = ColorTable::colorStrToColorRef(value);
		logBrush.lbStyle = BS_SOLID;

		HBRUSH	hbkBrush = ::CreateBrushIndirect(&logBrush);
		if (hbkBrush == NULL)
		{
			DEBUG_OUTPUT(("Failed to create brush in DlgFrame::setExProperty(%s, %s)", (LPCSTR)key, (LPCSTR)value));
			return 0;
		}

		m_crefBkColor = logBrush.lbColor;
		if (m_hbkBrush != NULL)
		{
			::DeleteObject(m_hbkBrush);
		}
		m_hbkBrush = hbkBrush;
	}

	if (m_hwndFrame != NULL)
	{
//		::InvalidateRect(m_hwndFrame, NULL, TRUE);
//		::UpdateWindow(m_hwndFrame);
		::RedrawWindow(m_hwndFrame, NULL, NULL, RDW_ERASE | RDW_INVALIDATE | RDW_ALLCHILDREN | RDW_UPDATENOW);
	}

	DEBUG_OUTPUT(("Leave DlgFrame::setExProperty(%s, %s)", (LPCSTR)key, (LPCSTR)value));

	return 1;
}

StringBuffer
DlgFrame::getExProperty(const StringBuffer& key)
{
	if (key.compareTo("bgcolor") == 0)
	{
		if (m_hbkBrush != NULL)
		{
			return ColorTable::colorRefToColorStr(m_crefBkColor);
		}
		else
		{
			return ColorTable::colorRefToColorStr(::GetSysColor(COLOR_3DFACE));
		}
	}

	return nullStr;
}

